#
# PyOtherSide: Asynchronous Python 3 Bindings for Qt 5 and Qt 6
# Copyright (c) 2011, 2013, Thomas Perl <m@thp.io>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

import matplotlib.pyplot as plt
import numpy as np
import io

import pyotherside

def plot_provider(image_id, requested_size):
    plt.figure()
    x = np.arange(0.0, 5.0, 0.1)
    plt.plot(np.sin(x))
    for dataset in image_id.split(';'):
        plt.plot([float(x) for x in dataset.split(',') if x])
    plt.title("Dataset from QML")
    buf = io.BytesIO()
    plt.savefig(buf, format='png')
    plt.close()
    buf.seek(0)
    return bytearray(buf.read()), requested_size, pyotherside.format_data

pyotherside.set_image_provider(plot_provider)



# Example how to fill a list model with data from Python.
def get_data():
    return [
        {'name': 'Alpha', 'team': 'red'},
        {'name': 'Beta', 'team': 'blue'},
        {'name': 'Gamma', 'team': 'green'},
        {'name': 'Delta', 'team': 'yellow'},
        {'name': 'Epsilon', 'team': 'orange'},
    ]



#
# PyOtherSide: Asynchronous Python 3 Bindings for Qt 5 and Qt 6
# Copyright (c) 2011, 2013, Thomas Perl <m@thp.io>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

import pyotherside
import math

def render(image_id, requested_size):
    print('image_id: "{image_id}", size: {requested_size}'.format(**locals()))

    # width and height will be -1 if not set in QML
    if requested_size == (-1, -1):
        requested_size = (300, 300)

    width, height = requested_size

    # center for circle
    cx, cy = width/2, 10

    pixels = []
    for y in range(height):
        for x in range(width):
            pixels.extend(reversed([
                255, # alpha
                int(10 + 10 * ((x - y * 0.5) % 20)), # red
                20 + 10 * (y % 20), # green
                int(255 * abs(math.sin(0.3*math.sqrt((cx-x)**2 + (cy-y)**2)))) # blue
            ]))
    return bytearray(pixels), (width, height), pyotherside.format_argb32

pyotherside.set_image_provider(render)



#
# PyOtherSide: Asynchronous Python 3 Bindings for Qt 5 and Qt 6
# Copyright (c) 2011, 2013, Thomas Perl <m@thp.io>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

import pyotherside

import os

HERE = os.path.dirname(__file__) or '.'

def load_data(image_id, requested_size):
    # If you return data in the format "pyotherside.format_svg_data" requested_size
    # is is used as the target size when rendering the SVG image

    # We use the image id to get name of the SVG file to render
    with open(os.path.join(HERE, image_id), 'rb') as f:
        svg_image_data = f.read()

    return bytearray(svg_image_data), requested_size, pyotherside.format_svg_data

pyotherside.set_image_provider(load_data)



import pyotherside
import os.path
import sys

print('Hello from module!')
print(sys.path)
print('file exists?', pyotherside.qrc_is_file('qrc_example.qml'))
print('file exists?', pyotherside.qrc_is_file('qrc_example.qml.nonexistent'))
print('dir exists?', pyotherside.qrc_is_dir('/'))
print('dir exists?', pyotherside.qrc_is_dir('/nonexistent'))

print('='*30)
def walk(root):
    for entry in pyotherside.qrc_list_dir(root):
        name = os.path.join(root, entry)
        if pyotherside.qrc_is_dir(name):
            walk(name)
        else:
            print(name, '=', len(pyotherside.qrc_get_file_contents(name)), 'bytes')
walk('/')
print('='*30)
print(pyotherside.qrc_get_file_contents('qrc_example.py').decode('utf-8'))
print('='*30)

try:
    print('dir exists with number', pyotherside.qrc_is_dir(123))
except Exception as e:
    print('got exception (as expected):', e)

try:
    print('file exists with none', pyotherside.qrc_is_file(None))
except Exception as e:
    print('got exception (as expected):', e)

try:
    print('dir entries with invalid', pyotherside.qrc_list_dir('/nonexistent'))
except Exception as e:
    print('got exception (as expected):', e)

try:
    print('file contents with invalid', pyotherside.qrc_get_file_contents('/qrc_example.qml.nonexistent'))
except Exception as e:
    print('got exception (as expected):', e)



import struct

def get_bytes():
    return struct.pack('<bhi', 123, 456, 789)

def set_bytes(bytedata):
    print('Got this raw data from QML:', bytedata)
    a, b, c = struct.unpack('<hhI', bytedata)
    print('Got this data (unpacked) from QML:', a, b, hex(c))
    return a == 1337 and b == -4711 and c == 0xcafebabe



import numpy

from OpenGL.GL import *
from OpenGL.GL.shaders import compileShader, compileProgram

VERTEX_SHADER = """#version 120
attribute vec4 vertices;
varying vec2 coords;

void main() {
    gl_Position = vertices;
    coords = vertices.xy;
}
"""

FRAGMENT_SHADER = """#version 120
uniform float t;
varying vec2 coords;
void main() {
    float i = 1. - (pow(abs(coords.x), 4.) + pow(abs(coords.y), 4.));
    i = smoothstep(t - 0.8, t + 0.8, i);
    i = floor(i * 20.) / 20.;
    gl_FragColor = vec4(coords * .5 + .5, i, i);
}
"""

class Renderer(object):

    def __init__(self):
        self.t = 0.0
        self.values = numpy.array([
            -1.0, -1.0,
            1.0, -1.0,
            -1.0, 1.0,
            1.0, 1.0
        ], dtype=numpy.float32)

    def set_t(self, t):
        self.t = t

    def init(self):
        self.vertexbuffer = glGenBuffers(1)
        vertex_shader = compileShader(VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = compileShader(FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = compileProgram(vertex_shader, fragment_shader)
        self.vertices_attr = glGetAttribLocation(self.program, b'vertices')
        self.t_attr = glGetUniformLocation(self.program, b't')

    def reshape(self, x, y, width, height):
        glViewport(x, y, width, height)

    def render(self):
        glUseProgram(self.program)
        try:
            glDisable(GL_DEPTH_TEST)
            glClearColor(0, 0, 0, 1)
            glClear(GL_COLOR_BUFFER_BIT)
            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE)

            glBindBuffer(GL_ARRAY_BUFFER, self.vertexbuffer)
            glEnableVertexAttribArray(self.vertices_attr)
            glBufferData(GL_ARRAY_BUFFER, self.values, GL_STATIC_DRAW)
            glVertexAttribPointer(self.vertices_attr, 2, GL_FLOAT, GL_FALSE, 0, None)
            glUniform1f(self.t_attr, self.t)

            glDrawArrays(GL_TRIANGLE_STRIP, 0, 4)
        finally:
            glDisableVertexAttribArray(0)
            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glUseProgram(0)

    def cleanup(self):
        glDeleteProgram(self.program)
        glDeleteBuffers(1, [self.vertexbuffer])



#
# PyOtherSide: Asynchronous Python 3 Bindings for Qt 5 and Qt 6
# Copyright (c) 2014, Thomas Perl <m@thp.io>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

import sys
from importlib import abc
from importlib.util import spec_from_loader

import pyotherside


def get_filename(fullname):
    basename = fullname.replace(".", "/")

    for import_path in sys.path:
        if not import_path.startswith("qrc:"):
            continue

        for candidate in ("{}/{}.py", "{}/{}/__init__.py"):
            filename = candidate.format(import_path, basename)
            if pyotherside.qrc_is_file(filename[len("qrc:"):]):
                return filename


class PyOtherSideQtRCLoader(abc.SourceLoader):
    def __init__(self, filepath):
        self.filepath = filepath

    def get_data(self, path):
        return pyotherside.qrc_get_file_contents(self.filepath[len("qrc:"):])

    def get_filename(self, fullname):
        return get_filename(fullname)


class PyOtherSideQtRCImporter(abc.MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        if path is None or all(x.startswith('qrc:') for x in path):
            fname = get_filename(fullname)
            if fname:
                return spec_from_loader(fullname, PyOtherSideQtRCLoader(fname))
        return None


sys.meta_path.append(PyOtherSideQtRCImporter())



#
# PyOtherSide: Asynchronous Python 3 Bindings for Qt 5 and Qt 6
# Copyright (c) 2011, 2013, Thomas Perl <m@thp.io>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

import pyotherside

import os
import random

from PIL import Image
from PIL import ImageDraw

def render(image_id, requested_size):
    # width and height will be -1 if not set in QML
    if requested_size == (-1, -1):
        requested_size = (300, 300)

    img = Image.new('RGBA', requested_size)

    filename = os.path.join(os.path.dirname(__file__), image_id)
    logo = Image.open(filename)

    for x in range(100):
        img.paste(logo, (random.randint(0, requested_size[0]), random.randint(0, requested_size[1])))

    draw = ImageDraw.Draw(img)
    for x in range(100):
        draw.text((random.randint(0, requested_size[0]), random.randint(0, requested_size[1])),
                'PyOtherSide PIL Test',
                fill=(random.randint(10, 255), random.randint(10, 255), random.randint(10, 255)))
    del draw

    b, g, r, a = img.split()
    img = Image.merge("RGBA", (r, g, b, a))
    return bytearray(img.tobytes()), img.size, pyotherside.format_argb32

pyotherside.set_image_provider(render)



# This examples shows how to do cleanups and other things when
# the application exists by using pyside.atexit().

import pyotherside

def called_when_exiting():
    print('Now exiting the application...')

pyotherside.atexit(called_when_exiting)
print('python loaded')



# This example demonstrates the use of pyotherside.send() to send events to Qt.

import pyotherside

import threading
import time

print('Using PyOtherSide version', pyotherside.version)

COLORS = ['red', 'green', 'blue']

def thread_func():
    i = 0
    while True:
        pyotherside.send('append', 'Next Number: ',  i)
        if i % 2 == 0:
            color = COLORS[int((i / 2) % len(COLORS))]
            pyotherside.send('color', color)
        i += 1
        time.sleep(1)

thread = threading.Thread(target=thread_func)
thread.start()



import threading
import time
import pyotherside

# If you try to instantiate a QObject, it's unbound
unbound = pyotherside.QObject()
print(unbound)
try:
    unbound.a = 1
except Exception as e:
    print('Got exception:', e)

def do_something(bar):
    while True:
        print('got: ', bar.dynamicFunction(1, 2, 3))
        time.sleep(1)

def foo(bar, py):
    # Printing the objects will give some info on the
    # QObject class and memory address
    print('got:', bar, py)

    # Ok, this is pretty wicked - we can now call into
    # the PyOtherSide QML element from within Python
    # (not that it's a good idea to do this, mind you..)
    print(py.evaluate)
    print(py.evaluate('3*3'))

    try:
        bar.i_am_pretty_sure_this_attr_does_not_exist = 147
    except Exception as e:
        print('Got exception (as expected):', e)

    try:
        bar.x = 'i dont think i can set this to a string'
    except Exception as e:
        print('Got exception (as expected):', e)

    # This doesn't work yet, because we can't convert a bound
    # member function to a Qt/QML type yet (fallback to None)
    try:
        bar.dynamicFunction(bar.dynamicFunction, 2, 3)
    except Exception as e:
        print('Got exception (as expected):', e)

    # Property access works just like expected
    print(bar.x, bar.color, bar.scale)
    bar.x *= 3

    # Printing a member function gives a bound method
    print(bar.dynamicFunction)

    # Calling a member function is just as easy
    result = bar.dynamicFunction(1, 2, 3)
    print('result:', result)

    try:
        bar.dynamicFunction(1, 2, 3, unexpected=123)
    except Exception as e:
        print('Got exception (as expected):', e)

    threading.Thread(target=do_something, args=[bar]).start()

    # Returning QObject references from Python also works
    return bar
